# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import sys
import string
from random import choice, randint, randrange

PY3 = sys.version_info[0] == 3

def get_random_words(count):
    russian = 'абвгдеёжзиклмнопрстуфхцчъыьэюя'
    alphabet = russian + string.ascii_lowercase

    def random_word(length):
        return "".join([choice(alphabet) for x in range(randint(1, length))])

    return list(set(random_word(10) for y in range(count)))


def get_random_binary(count):

    if PY3:
        def random_data(length):
            return bytes(randint(0, 255) for x in range(randint(1, length)))
    else:
        def random_data(length):
            return b"".join([chr(randint(0, 255)) for x in range(randint(1, length))])

    return list(set(random_data(10) for y in range(count)))



cdef extern from "../lib/marisa/key.h" namespace "marisa":

    cdef cppclass Key:
        Key() nogil
        Key(Key &query) nogil

        #Key &operator=(Key &query)

        char operator[](int i) nogil

        void set_str(char *str) nogil
        void set_str(char *ptr, int length) nogil
        void set_id(int id) nogil
        void set_weight(float weight) nogil

        char *ptr() nogil
        int length() nogil
        int id() nogil
        float weight() nogil

        void clear() nogil
        void swap(Key &rhs) nogil


from std_iostream cimport istream, ostream
from trie cimport Trie

cdef extern from "../lib/marisa/iostream.h" namespace "marisa":

    istream &read(istream &stream, Trie *trie) nogil
    ostream &write(ostream &stream, Trie &trie) nogil



cdef extern from "../lib/marisa/base.h":

    # A dictionary consists of 3 tries in default. Usually more tries make a
    # dictionary space-efficient but time-inefficient.
    ctypedef enum marisa_num_tries:
        MARISA_MIN_NUM_TRIES
        MARISA_MAX_NUM_TRIES
        MARISA_DEFAULT_NUM_TRIES


    # This library uses a cache technique to accelerate search functions. The
    # following enumerated type marisa_cache_level gives a list of available cache
    # size options. A larger cache enables faster search but takes a more space.
    ctypedef enum marisa_cache_level:
        MARISA_HUGE_CACHE
        MARISA_LARGE_CACHE
        MARISA_NORMAL_CACHE
        MARISA_SMALL_CACHE
        MARISA_TINY_CACHE
        MARISA_DEFAULT_CACHE

    # This library provides 2 kinds of TAIL implementations.
    ctypedef enum marisa_tail_mode:
        # MARISA_TEXT_TAIL merges last labels as zero-terminated strings. So, it is
        # available if and only if the last labels do not contain a NULL character.
        # If MARISA_TEXT_TAIL is specified and a NULL character exists in the last
        # labels, the setting is automatically switched to MARISA_BINARY_TAIL.
        MARISA_TEXT_TAIL

        # MARISA_BINARY_TAIL also merges last labels but as byte sequences. It uses
        # a bit vector to detect the end of a sequence, instead of NULL characters.
        # So, MARISA_BINARY_TAIL requires a larger space if the average length of
        # labels is greater than 8.
        MARISA_BINARY_TAIL

        MARISA_DEFAULT_TAIL

    # The arrangement of nodes affects the time cost of matching and the order of
    # predictive search.
    ctypedef enum marisa_node_order:
        # MARISA_LABEL_ORDER arranges nodes in ascending label order.
        # MARISA_LABEL_ORDER is useful if an application needs to predict keys in
        # label order.
        MARISA_LABEL_ORDER

        # MARISA_WEIGHT_ORDER arranges nodes in descending weight order.
        # MARISA_WEIGHT_ORDER is generally a better choice because it enables faster
        # matching.
        MARISA_WEIGHT_ORDER
        MARISA_DEFAULT_ORDER

    ctypedef enum marisa_config_mask:
        MARISA_NUM_TRIES_MASK
        MARISA_CACHE_LEVEL_MASK
        MARISA_TAIL_MODE_MASK
        MARISA_NODE_ORDER_MASK
        MARISA_CONFIG_MASK






cimport query, key

cdef extern from "../lib/marisa/keyset.h" namespace "marisa":
    cdef cppclass Keyset:

#        cdef enum constants:
#            BASE_BLOCK_SIZE  = 4096
#            EXTRA_BLOCK_SIZE = 1024
#            KEY_BLOCK_SIZE   = 256

        Keyset() nogil

        void push_back(key.Key &key) nogil
        void push_back(key.Key &key, char end_marker) nogil

        void push_back(char *str) nogil
        void push_back(char *ptr, int length) nogil
        void push_back(char *ptr, int length, float weight) nogil

        key.Key &operator[](int i) nogil

        int num_keys() nogil
        bint empty() nogil

        int size() nogil
        int total_length() nogil

        void reset() nogil
        void clear() nogil
        void swap(Keyset &rhs) nogil



from libcpp.string cimport string

cdef extern from "<istream>" namespace "std":
    cdef cppclass istream:
        istream() nogil except +
        istream& read (char* s, int n) nogil except +

    cdef cppclass ostream:
        ostream() nogil except +
        ostream& write (char* s, int n) nogil except +

cdef extern from "<sstream>" namespace "std":

    cdef cppclass stringstream:
        stringstream() nogil
        stringstream(string s) nogil
        string str () nogil




cimport query, key

cdef extern from "../lib/marisa/agent.h" namespace "marisa":
    cdef cppclass Agent:
        Agent() except +

        query.Query &query() nogil
        key.Key &key() nogil

        void set_query(char *str) nogil
        void set_query(char *ptr, int length) nogil
        void set_query(int key_id) nogil

        void set_key(char *str) nogil
        void set_key(char *ptr, int length) nogil
        void set_key(int id) nogil

        void clear() nogil

        void init_state() nogil

        void swap(Agent &rhs) nogil




#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals, division
import random
import string
import timeit
import os
import zipfile
import struct
#import pstats
#import cProfile

import marisa_trie

def words100k():
    zip_name = os.path.join(
        os.path.abspath(os.path.dirname(__file__)),
        'words100k.txt.zip'
    )
    zf = zipfile.ZipFile(zip_name)
    txt = zf.open(zf.namelist()[0]).read().decode('utf8')
    return txt.splitlines()

def random_words(num):
    russian = 'абвгдеёжзиклмнопрстуфхцчъыьэюя'
    alphabet = '%s%s' % (russian, string.ascii_letters)
    return [
        "".join([random.choice(alphabet) for x in range(random.randint(1,15))])
        for y in range(num)
    ]

def truncated_words(words):
    return [word[:3] for word in words]

def prefixes1k(words, prefix_len):
    words = [w for w in words if len(w) >= prefix_len]
    every_nth = int(len(words)/1000)
    _words = [w[:prefix_len] for w in words[::every_nth]]
    return _words[:1000]

WORDS100k = words100k()
MIXED_WORDS100k = truncated_words(WORDS100k)
NON_WORDS100k = random_words(100000)
PREFIXES_3_1k = prefixes1k(WORDS100k, 3)
PREFIXES_5_1k = prefixes1k(WORDS100k, 5)
PREFIXES_8_1k = prefixes1k(WORDS100k, 8)
PREFIXES_15_1k = prefixes1k(WORDS100k, 15)


def format_result(key, value):
    print("%55s:    %s" % (key, value))


def bench(name, timer, descr='M ops/sec', op_count=0.1, repeats=3, runs=5):
    times = []
    for x in range(runs):
        times.append(timer.timeit(repeats))

    def op_time(time):
        return op_count*repeats / time

    val = "%0.3f%s" % (op_time(min(times)), descr)
    format_result(name, val)

def create_trie():
    words = words100k()
    return marisa_trie.Trie(words)

def create_bytes_trie():
    words = words100k()
    values = [struct.pack(str('<H'), len(word)) for word in words]
    return marisa_trie.BytesTrie(zip(words, values))

def create_record_trie():
    words = words100k()
    values = [ [len(word)] for word in words]
    return marisa_trie.RecordTrie(str('<H'), zip(words, values))


def benchmark():
    print('\n====== Benchmarks (100k unique unicode words) =======\n')

    tests = [
        ('__getitem__ (hits)', "for word in WORDS100k: data[word]", 'M ops/sec', 0.1, 3),
        ('__contains__ (hits)', "for word in WORDS100k: word in data", 'M ops/sec', 0.1, 3),
        ('__contains__ (misses)', "for word in NON_WORDS100k: word in data", 'M ops/sec', 0.1, 3),
        #('__setitem__ (updates)', 'for word in WORDS100k: data[word]=1', 'M ops/sec',0.1, 3),
        #('__setitem__ (inserts)', 'for word in NON_WORDS_10k: data[word]=1', 'M ops/sec',0.01, 3),
        #('setdefault (updates)', 'for word in WORDS100k: data.setdefault(word, 1)', 'M ops/sec', 0.1, 3),
        #('setdefault (inserts)', 'for word in  NON_WORDS_10k: data.setdefault(word, 1)', 'M ops/sec', 0.01, 3),
        ('items()', 'list(data.items())', ' ops/sec', 1, 1),
        ('keys()', 'list(data.keys())', ' ops/sec', 1, 1),
#        ('values()', 'list(data.values())', ' ops/sec', 1, 1),
    ]

    common_setup = """
from __main__ import create_trie, create_bytes_trie, create_record_trie
from __main__ import WORDS100k, NON_WORDS100k, MIXED_WORDS100k
from __main__ import PREFIXES_3_1k, PREFIXES_5_1k, PREFIXES_8_1k, PREFIXES_15_1k
NON_WORDS_10k = NON_WORDS100k[:10000]
NON_WORDS_1k = ['ыва', 'xyz', 'соы', 'Axx', 'avы']*200
"""
    dict_setup = common_setup + 'data = dict((word, len(word)) for word in WORDS100k);'
    trie_setup = common_setup + 'data = create_trie();'
    bytes_trie_setup = common_setup + 'data = create_bytes_trie();'
    record_trie_setup = common_setup + 'data = create_record_trie();'

    structures = [
        ('dict', dict_setup),
        ('Trie', trie_setup),
        ('BytesTrie', bytes_trie_setup),
        ('RecordTrie', record_trie_setup),
    ]
    for test_name, test, descr, op_count, repeats in tests:
        for name, setup in structures:
            timer = timeit.Timer(test, setup)
            full_test_name = "%s %s" % (name, test_name)
            try:
                bench(full_test_name, timer, descr, op_count, repeats)
            except (TypeError, AttributeError):
                format_result(full_test_name, "not supported")


    # trie-specific benchmarks
    _bench_data = [
        ('hits', 'WORDS100k'),
        ('mixed', 'MIXED_WORDS100k'),
        ('misses', 'NON_WORDS100k'),
    ]

    for meth in ['prefixes', 'iter_prefixes']:
        for name, data in _bench_data:
            bench(
                'Trie.%s (%s)' % (meth, name),
                timeit.Timer(
                    "for word in %s:\n"
                    "   for it in data.%s(word): pass" % (data, meth),
                    trie_setup
                )
            )

    _bench_data = [
        ('xxx', 'avg_len(res)==415', 'PREFIXES_3_1k'),
        ('xxxxx', 'avg_len(res)==17', 'PREFIXES_5_1k'),
        ('xxxxxxxx', 'avg_len(res)==3', 'PREFIXES_8_1k'),
        ('xxxxx..xx', 'avg_len(res)==1.4', 'PREFIXES_15_1k'),
        ('xxx', 'NON_EXISTING', 'NON_WORDS_1k'),
    ]
    for xxx, avg, data in _bench_data:
        for meth in ['keys']: #('items', 'keys', 'values'):
            bench(
                'Trie.%s(prefix="%s"), %s' % (meth, xxx, avg),
                timeit.Timer(
                    "for word in %s: data.%s(word)" % (data, meth),
                    trie_setup
                ),
                'K ops/sec',
                op_count=1,
            )

def check_trie(trie, words):
    value = 0
    for word in words:
        value += trie[word]
    if value != len(words):
        raise Exception()

def profiling():
    import pstats
    import cProfile
    print('\n====== Profiling =======\n')
    trie = create_trie()
    WORDS = words100k()

#    def check_prefixes(trie, words):
#        for word in words:
#            trie.keys(word)
#    cProfile.runctx("check_prefixes(trie, NON_WORDS_1k)", globals(), locals(), "Profile.prof")
#
    cProfile.runctx("check_trie(trie, WORDS)", globals(), locals(), "Profile.prof")

    s = pstats.Stats("Profile.prof")
    s.strip_dirs().sort_stats("time").print_stats(20)


if __name__ == '__main__':
    benchmark()
    #profiling()
    print('\n~~~~~~~~~~~~~~\n')


# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals
import tempfile
import pickle

import pytest

import marisa_trie
from .utils import get_random_words


def test_build():
    keys = get_random_words(1000)
    trie = marisa_trie.Trie(keys)

    for key in keys:
        assert key in trie

    non_key = '2135'
    assert non_key not in trie


def test_contains():
    trie = marisa_trie.Trie()
    assert 'foo' not in trie

    trie = marisa_trie.Trie(['foo'])
    assert 'foo' in trie
    assert 'f' not in trie

def test_key_id():
    words = ['foo', 'bar', 'f']
    trie = marisa_trie.Trie(words)
    for word in words:
        key_id = trie.key_id(word)
        assert trie.restore_key(key_id) == word


    key_ids = [trie.key_id(word) for word in words]
    non_existing_id = max(key_ids) + 1

    with pytest.raises(KeyError):
        trie.restore_key(non_existing_id)

    with pytest.raises(KeyError):
        print(trie.key_id('fo'))

def test_saveload():
    fd, fname = tempfile.mkstemp()

    words = ['foo', 'bar', 'f']
    trie = marisa_trie.Trie(words)

    with open(fname, 'w') as f:
        trie.write(f)

    with open(fname, 'r') as f:
        trie2 = marisa_trie.Trie()
        trie2.read(f)

    for word in words:
        assert word in trie2


def test_mmap():
    fd, fname = tempfile.mkstemp()
    words = get_random_words(1000)
    trie = marisa_trie.Trie(words)
    with open(fname, 'w') as f:
        trie.write(f)

    trie2 = marisa_trie.Trie()
    trie2.mmap(fname)

    for word in words:
        assert word in trie2

def test_dumps_loads():
    words = get_random_words(1000)
    trie = marisa_trie.Trie(words)
    data = trie.tobytes()

    trie2 = marisa_trie.Trie().frombytes(data)

    for word in words:
        assert word in trie2
        assert trie2.key_id(word) == trie.key_id(word)

def test_pickling():
    words = get_random_words(1000)
    trie = marisa_trie.Trie(words)

    data = pickle.dumps(trie)
    trie2 = pickle.loads(data)

    for word in words:
        assert word in trie2
        assert trie2.key_id(word) == trie.key_id(word)


def test_len():
    trie = marisa_trie.Trie()
    assert len(trie) == 0

    trie = marisa_trie.Trie(['foo', 'f', 'bar'])
    assert len(trie) == 3

def test_prefixes():
    trie = marisa_trie.Trie(['foo', 'f', 'foobar', 'bar'])
    assert trie.prefixes('foobar') == ['f', 'foo', 'foobar']
    assert trie.prefixes('foo') == ['f', 'foo']
    assert trie.prefixes('bar') == ['bar']
    assert trie.prefixes('b') == []

    assert list(trie.iter_prefixes('foobar')) == ['f', 'foo', 'foobar']

def test_keys():
    keys = ['foo', 'f', 'foobar', 'bar']
    trie = marisa_trie.Trie(keys)
    assert set(trie.keys()) == set(keys)

def test_keys_prefix():
    keys = ['foo', 'f', 'foobar', 'bar']
    trie = marisa_trie.Trie(keys)
    assert set(trie.keys('fo')) == set(['foo', 'foobar'])
    assert trie.keys('foobarz') == []


#def test_int_trie():
#    data = {'foo': 10, 'f': 20, 'foobar':30, 'bar': 40}
#    trie = marisa_trie.IntTrie().build(data)
#
#    for key, value in data.items():
#        assert trie[key] == value
#
#    trie['foo'] = 500
#    assert trie['foo'] == 500
#
#    # trie is frozen, no new keys are allowed
#    with pytest.raises(KeyError):
#        trie['z'] = 200


cdef extern from "../lib/marisa/query.h" namespace "marisa":

    cdef cppclass Query:
        Query() nogil
        Query(Query &query) nogil

        #Query &operator=(Query &query)

        char operator[](int i) nogil

        void set_str(char *str) nogil
        void set_str(char *ptr, int length) nogil
        void set_id(int id) nogil

        char *ptr() nogil
        int length() nogil
        int id() nogil

        void clear() nogil
        void swap(Query &rhs) nogil


from __future__ import unicode_literals
#from cython cimport view
#cimport cpython.array
#import array
#from libc.stdlib cimport free, malloc

from std_iostream cimport stringstream, istream, ostream
cimport keyset
cimport key
cimport query
cimport agent
cimport trie
cimport iostream
cimport base

import struct


DEFAULT_CACHE = base.MARISA_DEFAULT_CACHE
HUGE_CACHE = base.MARISA_HUGE_CACHE
LARGE_CACHE = base.MARISA_LARGE_CACHE
NORMAL_CACHE = base.MARISA_NORMAL_CACHE
SMALL_CACHE = base.MARISA_SMALL_CACHE
TINY_CACHE = base.MARISA_TINY_CACHE
DEFAULT_CACHE = base.MARISA_DEFAULT_CACHE

MIN_NUM_TRIES = base.MARISA_MIN_NUM_TRIES
MAX_NUM_TRIES = base.MARISA_MAX_NUM_TRIES
DEFAULT_NUM_TRIES = base.MARISA_DEFAULT_NUM_TRIES

# MARISA_TEXT_TAIL merges last labels as zero-terminated strings. So, it is
# available if and only if the last labels do not contain a NULL character.
# If MARISA_TEXT_TAIL is specified and a NULL character exists in the last
# labels, the setting is automatically switched to MARISA_BINARY_TAIL.
TEXT_TAIL = base.MARISA_TEXT_TAIL

# MARISA_BINARY_TAIL also merges last labels but as byte sequences. It uses
# a bit vector to detect the end of a sequence, instead of NULL characters.
# So, MARISA_BINARY_TAIL requires a larger space if the average length of
# labels is greater than 8.
BINARY_TAIL = base.MARISA_BINARY_TAIL
DEFAULT_TAIL = base.MARISA_DEFAULT_TAIL


# MARISA_LABEL_ORDER arranges nodes in ascending label order.
# MARISA_LABEL_ORDER is useful if an application needs to predict keys in
# label order.
LABEL_ORDER = base.MARISA_LABEL_ORDER

# MARISA_WEIGHT_ORDER arranges nodes in descending weight order.
# MARISA_WEIGHT_ORDER is generally a better choice because it enables faster
# matching.
WEIGHT_ORDER = base.MARISA_WEIGHT_ORDER
DEFAULT_ORDER = base.MARISA_DEFAULT_ORDER


cdef class _Trie:
    """
    Base MARISA-trie wrapper.
    It can store unicode keys and assigns an unque ID to each key.
    """

    cdef trie.Trie* _trie

    def __init__(self, arg=None, num_tries=DEFAULT_NUM_TRIES, binary=False,
                        cache_size=DEFAULT_CACHE, order=DEFAULT_ORDER):
        """
        ``arg`` must be an iterable with unicode keys or None
        if you're going to load a trie later.
        """

        if self._trie:
            return
        self._trie = new trie.Trie()

        byte_keys = (key.encode('utf8') for key in sorted(arg or []))
        self._build(
            byte_keys,
            num_tries=num_tries,
            binary=binary,
            cache_size=cache_size,
            order=order
        )


    def __dealloc__(self):
        if self._trie:
            del self._trie


    def _config_flags(self, num_tries=DEFAULT_NUM_TRIES, binary=False,
                            cache_size=DEFAULT_CACHE, order=DEFAULT_ORDER):

        if not MIN_NUM_TRIES <= num_tries <= MAX_NUM_TRIES:
            raise ValueError("num_tries (which is %d) must be between between %d and %d" % (num_tries, MIN_NUM_TRIES, MAX_NUM_TRIES))

        binary_flag = BINARY_TAIL if binary else TEXT_TAIL
        return num_tries | binary_flag | cache_size | order


    def _build(self, byte_keys, **options):
        """
        Builds the trie using values from ``byte_keys`` iterable.
        """
        cdef char* data
        cdef keyset.Keyset *ks = new keyset.Keyset()

        try:
            for key in byte_keys:
                data = key
                ks.push_back(data, len(key))
            self._trie.build(ks[0], self._config_flags(**options))
        finally:
            del ks

    def __len__(self):
        return self._trie.num_keys()

    def __contains__(self, unicode key):
        cdef bytes _key = key.encode('utf8')
        return self._contains(_key)

    cdef bint _contains(self, bytes key):
        cdef agent.Agent ag
        ag.set_query(key)
        return self._trie.lookup(ag)

    def read(self, f):
        """
        Reads a trie from an open file object.

        Works only with "real" disk-based file objects,
        file-like objects are not supported.
        """
        self._trie.read(f.fileno())

    def write(self, f):
        """
        Reads a trie to an open file object.

        Works only with "real" disk-based file objects,
        file-like objects are not supported.
        """
        self._trie.write(f.fileno())

    def save(self, path):
        """ Saves trie to a file. """
        with open(path, 'w') as f:
            self.write(f)

    def load(self, path):
        """ Loads trie from a file. """
        with open(path, 'r') as f:
            self.read(f)

    cpdef bytes tobytes(self) except +:
        """
        Returns raw trie content as bytes.
        """
        cdef stringstream stream
        iostream.write((<ostream *> &stream)[0], self._trie[0])
        cdef bytes res = stream.str()
        return res

    cpdef frombytes(self, bytes data) except +:
        """
        Loads trie from bytes ``data``.
        """
        cdef stringstream* stream = new stringstream(data)
        try:
            iostream.read((<istream *> stream)[0], self._trie)
        finally:
            del stream
        return self


    def __reduce__(self): # pickling support
        return self.__class__, tuple(), self.tobytes()

    def __setstate__(self, state): # pickling support
        self.frombytes(state)


    def mmap(self, path):
        """
        Mmaps trie to a file; this allows lookups without loading full
        trie to memory.
        """
        import sys
        str_path = path.encode(sys.getfilesystemencoding())
        cdef char* c_path = str_path
        self._trie.mmap(c_path)

    cpdef list keys(self, unicode prefix=""):
        """
        Returns a list with all keys with a prefix ``prefix``.
        """
        # non-generator inlined version of iterkeys()
        cdef list res = []
        cdef bytes b_key

        cdef bytes b_prefix = prefix.encode('utf8')
        cdef agent.Agent ag
        ag.set_query(b_prefix)

        while self._trie.predictive_search(ag):
            b_key = ag.key().ptr()[:ag.key().length()]
            res.append(b_key.decode('utf8'))

        return res


cdef class Trie(_Trie):
     """
     This trie stores unicode keys and assigns an unque ID to each key.
     """

     cpdef int key_id(self, unicode key) except -1:
         """
         Returns unique auto-generated key index for a ``key``.
         Raises KeyError if key is not in this trie.
         """
         cdef bytes _key = key.encode('utf8')
         cdef int res = self._key_id(_key)
         if res == -1:
             raise KeyError(key)
         return res

     cpdef unicode restore_key(self, int index):
         """
         Returns a key given its index (obtained by ``key_id`` method).
         """
         cdef agent.Agent ag
         ag.set_query(index)
         try:
             self._trie.reverse_lookup(ag)
         except KeyError:
             raise KeyError(index)
         cdef bytes _key = ag.key().ptr()
         return _key.decode('utf8')

     cdef int _key_id(self, char* key):
         cdef bint res
         cdef agent.Agent ag
         ag.set_query(key)
         res = self._trie.lookup(ag)
         if not res:
             return -1
         return ag.key().id()

     def iter_prefixes(self, unicode key):
         """
         Returns an iterator of all prefixes of a given key.
         """
         cdef agent.Agent ag
         cdef bytes b_prefix

         cdef bytes b_key = key.encode('utf8')
         ag.set_query(b_key)

         while self._trie.common_prefix_search(ag):
             b_prefix = ag.key().ptr()[:ag.key().length()]
             yield b_prefix.decode('utf8')

     def prefixes(self, unicode key):
         """
         Returns a list with all prefixes of a given key.
         """

         # this an inlined version of ``list(self.iter_prefixes(key))``

         cdef agent.Agent ag
         cdef bytes b_prefix
         cdef list res = []

         cdef bytes b_key = key.encode('utf8')
         ag.set_query(b_key)

         while self._trie.common_prefix_search(ag):
             b_prefix = ag.key().ptr()[:ag.key().length()]
             res.append(b_prefix.decode('utf8'))
         return res

     def iterkeys(self, unicode prefix=""):
         """
         Returns an iterator over keys that have a prefix ``prefix``.
         """
         cdef agent.Agent ag
         cdef bytes b_key
         cdef bytes b_prefix = prefix.encode('utf8')
         ag.set_query(b_prefix)

         while self._trie.predictive_search(ag):
             b_key = ag.key().ptr()[:ag.key().length()]
             yield b_key.decode('utf8')



# This symbol is not allowed in utf8 so it is safe to use
# as a separator between utf8-encoded string and binary payload.
DEF _VALUE_SEPARATOR = b'\xff'

cdef class BytesTrie(_Trie):
    """
    This class implements read-only Trie-based
    {unicode -> list of bytes objects} mapping.

    This mapping is implemented by appending binary values to
    utf8-encoded keys and storing the result in MARISA-trie.
    """

    def __init__(self, arg=None, **options):
        """
        ``arg`` must be an iterable of tuples (unicode_key, bytes_payload).
        """
        super(BytesTrie, self).__init__()
        if arg is None:
            arg = []
        byte_keys = (self._raw_key(d[0], d[1]) for d in sorted(arg))
        self._build(byte_keys, **options)


    cpdef bytes _raw_key(self, unicode key, bytes payload):
        return key.encode('utf8') + _VALUE_SEPARATOR + payload

    cdef bint _contains(self, bytes key):
        cdef agent.Agent ag
        cdef bytes _key = key + _VALUE_SEPARATOR
        ag.set_query(_key)
        return self._trie.predictive_search(ag)


    def __getitem__(self, key):
        cdef list res

        if isinstance(key, unicode):
            res = self.get_value(key)
        else:
            res = self.b_get_value(key)

        if not res:
            raise KeyError(key)
        return res


    cpdef list get_value(self, unicode key):
        """
        Returns a list of payloads (as byte objects) for a given unicode key.
        """
        cdef bytes b_key = key.encode('utf8')
        return self.b_get_value(b_key)


    cpdef list b_get_value(self, bytes key):
        """
        Returns a list of payloads (as byte objects) for a given utf8-encoded key.
        """
        cdef list res = []
        cdef bytes value
        cdef bytes b_prefix = key + _VALUE_SEPARATOR
        cdef int prefix_len = len(b_prefix)

        cdef agent.Agent ag
        ag.set_query(b_prefix)

        while self._trie.predictive_search(ag):
            value = ag.key().ptr()[prefix_len:ag.key().length()]
            res.append(value)

        return res

    cpdef list items(self, unicode prefix=""):
        cdef bytes b_prefix = prefix.encode('utf8')
        cdef bytes key, raw_key, value
        cdef list res = []

        cdef agent.Agent ag
        ag.set_query(b_prefix)

        while self._trie.predictive_search(ag):
            raw_key = ag.key().ptr()[:ag.key().length()]
            key, value = raw_key.split(_VALUE_SEPARATOR, 1)
            res.append(
                (key.decode('utf8'), value)
            )
        return res

    cpdef list keys(self, unicode prefix=""):
        items = self.items(prefix)
        if not items:
            return []
        keys, values = zip(*items)
        return list(keys)



cdef class RecordTrie(BytesTrie):
    """
    This class implements read-only Trie-based
    {unicode -> list of tuples} mapping where all tuples are of the
    same structure and may be packed with the same format string
    using python ``struct`` module from the standard library.

    The payload format must be defined at creation time using ``fmt``
    constructor argument; it has the same meaning as ``fmt`` argument
    for functions from ``struct`` module; take a look at
    http://docs.python.org/library/struct.html#format-strings for the
    specification.

    This mapping is implemented by appending binary values to
    utf8-encoded keys and storing the result in MARISA-trie.
    """
    cdef _struct

    def __init__(self, fmt, arg=None, **options):
        """
        ``arg`` must be an iterable of tuples (unicode_key, data_tuple).
        Data tuples will be converted to bytes with
        ``struct.pack(fmt, *data_tuple)``.

        Take a look at
        http://docs.python.org/library/struct.html#format-strings for the
        format string specification.
        """
        self._struct = struct.Struct(str(fmt))

        keys = ((d[0], self._struct.pack(*d[1])) for d in (arg or []))
        super(RecordTrie, self).__init__(keys, **options)


    cpdef list b_get_value(self, bytes key):
        cdef list values = BytesTrie.b_get_value(self, key)
        return [self._struct.unpack(val) for val in values]


    cpdef list items(self, unicode prefix=""):
        cdef list items = BytesTrie.items(self, prefix)
        return [(key, self._struct.unpack(val)) for (key, val) in items]



# -*- coding: utf-8 -*-
from __future__ import absolute_import, unicode_literals

import pytest

import marisa_trie
from .utils import get_random_words, get_random_binary


class TestBytesTrie(object):

    def data(self):
        keys = get_random_words(10000)
        values = [key.encode('cp1251') for key in keys]
        return list(zip(keys, values))

    def test_contains(self):
        data = self.data()
        trie = marisa_trie.BytesTrie(data)

        for key, value in data:
            assert key in trie

        non_key = '2135'
        assert non_key not in trie


    def test_getitem_fuzzy(self):
        data = self.data()
        trie = marisa_trie.BytesTrie(data)

        for key, value in data:
            assert trie[key] == [value]

        with pytest.raises(KeyError):
            trie['2135']

    def test_getitem_multiple(self):
        data = [
            ('foo', b'x'),
            ('fo',  b'y'),
            ('foo', b'a'),
        ]
        trie = marisa_trie.BytesTrie(data)
        assert trie['fo'] == [b'y']
        assert trie['foo'] == [b'a', b'x']

    def test_null_bytes_in_values(self):
        keys = get_random_words(10000)
        values = get_random_binary(10000)

        assert any(b'\x00' in p for p in values)

        data = zip(keys, values)
        trie = marisa_trie.BytesTrie(data)

        for key, value in data:
            assert trie[key] == [value]


    def test_keys(self):
        trie = marisa_trie.BytesTrie([
            ('foo', b'x'),
            ('fo',  b'y'),
            ('foo', b'a'),
        ])

        # FIXME: ordering?
        assert trie.keys() == ['foo', 'foo', 'fo']
        assert trie.keys('f') == ['foo', 'foo', 'fo']
        assert trie.keys('fo') == ['foo', 'foo', 'fo']
        assert trie.keys('foo') == ['foo', 'foo']
        assert trie.keys('food') == []
        assert trie.keys('bar') == []

    def test_items(self):
        data = [
            ('fo',  b'y'),
            ('foo', b'x'),
            ('foo', b'a'),
        ]
        trie = marisa_trie.BytesTrie(data)
        assert set(trie.items()) == set(data)
        assert set(trie.items('f')) == set(data)
        assert set(trie.items('fo')) == set(data)
        assert set(trie.items('foo')) == set(data[1:])
        assert trie.items('food') == []
        assert trie.items('bar') == []


class TestRecordTrie(object):

    def data(self):
        keys = get_random_words(10000)
        values = [(len(key), 'ё' in key) for key in keys]
        fmt = str("<H?")
        return fmt, list(zip(keys, values))


    def test_getitem(self):
        fmt, data = self.data()
        trie = marisa_trie.RecordTrie(fmt, data)

        for key, value in data:
            assert trie[key] == [value]

        with pytest.raises(KeyError):
            trie['2135']

    def test_items(self):
        fmt, data = self.data()
        trie = marisa_trie.RecordTrie(fmt, data)

        assert set(trie.items()) == set(data)
